#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun May  5 22:03:32 2019

@author: jkl
"""
EMagPy_version = '0.0.4'
name = 'emagpy'
from emagpy.Problem import Problem
from emagpy.Survey import Survey
